import requests
import csv
import re
import sqlite3
import logging
import pkg_resources

logger = logging.getLogger(__name__)

class ouidb:
    def __init__(self, url, raw_data_file, pipe_file, database_file):
        self.url = url
        self.raw_data_file = pkg_resources.resource_filename('pxgrid_pyshark', raw_data_file)
        self.pipe_file = pkg_resources.resource_filename('pxgrid_pyshark', pipe_file)
        self.database_file = pkg_resources.resource_filename('pxgrid_pyshark', database_file)
        self._initialize_database()

    def _initialize_database(self):
        self.download_macoui_data()
        self.create_pipe_separated_file()
        self.import_to_sqlite()

    def download_macoui_data(self):
        logger.debug('downloading OUI db - starting')
        response = requests.get(self.url)
        with open(self.raw_data_file, 'wb') as f:
            f.write(response.content)
        logger.debug('downloading OUI db - complete')

    def create_pipe_separated_file(self):
        logger.debug('parsing OUI db data - starting')
        with open(self.raw_data_file, 'r') as infile, open(self.pipe_file, 'w') as outfile:
            outfile.write("OrgName|OUI\n")
            for line in infile:
                line = line.rstrip('\n')
                if match := re.match(r'^([0-9A-F]{2})-([0-9A-F]{2})-([0-9A-F]{2})\s+\(hex\)\s+(.+)$', line):
                    orgname, oui = match.group(4), match.group(1) + match.group(2) + match.group(3)
                    outfile.write(f"{orgname}|{oui}\n")
        logger.debug('parsing OUI db data - complete')

    def import_to_sqlite(self):
        logger.debug('create OUI db table - starting')
        connection = sqlite3.connect(self.database_file)
        cursor = connection.cursor()

        # Delete any pre-existing DB data
        cursor.execute('DROP TABLE IF EXISTS macoui')

        # Create the macoui table
        cursor.execute('''
            CREATE TABLE IF NOT EXISTS macoui (
                OrgName TEXT,
                OUI TEXT
            )
        ''')

        # Import data into the macoui table
        with open(self.pipe_file, 'r') as f:
            reader = csv.reader(f, delimiter='|')
            next(reader)  # Skip header
            cursor.executemany('INSERT OR REPLACE INTO macoui (OrgName, OUI) VALUES (?, ?)', reader)

        # Add indices on each column
        cursor.execute('CREATE INDEX IF NOT EXISTS idx_oui ON macoui (OUI)')
        cursor.execute('CREATE INDEX IF NOT EXISTS idx_orgname ON macoui (OrgName)')

        # Commit changes and close connection
        connection.commit()
        connection.close()
        logger.debug('create OUI db table - complete')

    def query_mac_address(self, mac_address):
        connection = sqlite3.connect(self.database_file)
        cursor = connection.cursor()

        # Query the OrgName for the given MAC address
        cursor.execute('SELECT OrgName FROM macoui WHERE OUI = ?', (mac_address,))
        result = cursor.fetchone()
        connection.close()

        # Print the result
        if result:
            return result[0]
        else:
            return None


import pyshark
import logging
from pxgrid_pyshark import endpointsdb
from pxgrid_pyshark import parser

## DEDICATED CLASS FOR ALL PACKET PARSING FUNCTIONS

logger = logging.getLogger(__name__)
# parser = parser()

## Create dict of supported protocols and their appropriate inspection functions
packet_callbacks = {
    'sip': parser.parse_sip,
    'ssdp': parser.parse_ssdp,
    'mdns': parser.parse_mdns,
    'http': parser.parse_http,
    'xml': parser.parse_xml
}

class capture:
    ## Initialize endpoints database and parser
    def __init__(self):
        self.endpoints = endpointsdb()
        self.parser = parser()

    ## Process network packets using global Parser instance and dictionary of supported protocols
    def process_packet(self, packet):
        try:
            highest_layer = packet.highest_layer
            inspection_layer = str(highest_layer).split('_')[0]
            ## If XML traffic included over HTTP, match on XML parsing
            if inspection_layer == 'XML':
                fn = parser.parse_xml(packet)
                if fn is not None:
                    self.endpoints.update_db_list(fn)
            else:
                for layer in packet.layers:
                    fn = packet_callbacks.get(layer.layer_name)
                    if fn is not None:
                        self.endpoints.update_db_list(fn(packet))
        except Exception as e:
            logger.error(f'error processing packet details {highest_layer}: {e}')
    
    ##
    def capture_live_packets(self, network_interface, filter):
        try:
            logger.debug(f'beginning live capture of {network_interface}')
            capture = pyshark.LiveCapture(interface=network_interface, only_summaries=False, include_raw=True, use_json=True, display_filter=filter)
            currentPacket = 0
            for packet in capture.sniff_continuously():
                ## Wrap individual packet processing within 'try' statement to avoid formatting issues crashing entire process
                try:
                    self.process_packet(packet)
                except Exception as e:
                    logger.error(f'error processing packet: {currentPacket}: {e}')
                currentPacket += 1
        except Exception as e:
            logger.error(f'live capture process crashed: {e}')
            sys.exit(1)

    def process_capture_file(self, capture_file, capture_filter):
        if Path(capture_file).exists():
            logger.warning(f'processing capture file: {capture_file}')
            # start_time = time.perf_counter()
            capture = pyshark.FileCapture(capture_file, display_filter=capture_filter, only_summaries=False, include_raw=True, use_json=True)
            capture = pyshark.FileCapture()
            currentPacket = 0
            for packet in capture:
                ## Wrap individual packet processing within 'try' statement to avoid formatting issues crashing entire process
                try:
                    process_packet(packet)
                except Exception as e:
                    logger.error(f'Error processing packet: {capture_file}, packet # {currentPacket}: {e}')
                currentPacket += 1
            capture.close()
            # end_time = time.perf_counter()
            # logger.warning(f'processing capture file complete: execution time: {end_time - start_time:0.6f} : {currentPacket} packets processed ##')
        else:
            logger.error(f'capture file not found: {capture_file}')            



# This file was generated by 'versioneer.py' (0.29) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2024-03-13T15:19:05-0400",
 "dirty": false,
 "error": null,
 "full-revisionid": "0e67d198af1eec830c2baef7a0088f4583616c4d",
 "version": "0.0.8"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



import sqlite3
import logging
from datetime import datetime

logger = logging.getLogger(__name__)

records_updated = 0
records_created = 0
ise_updates = 0

class endpointsdb:
    def __init__(self, db_file='endpoint_database.db'):
        self.connection = sqlite3.connect(db_file)
        self.cursor = self.connection.cursor()

    def clear_database(self):
        self.cursor.execute('DROP TABLE IF EXISTS endpoints')
        logger.debug('clear endpooints DB - complete')

    ## Create table for holding endpoint data; remove table if already exists to avoid old data
    def create_database(self):
        logger.debug('create endpoints DB - starting')
        # self.cursor.execute('DROP TABLE IF EXISTS endpoints')
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS endpoints (
                mac TEXT PRIMARY KEY, protocol TEXT, ip TEXT,
                id TEXT, name TEXT, vendor TEXT, hw TEXT, sw TEXT, productID TEXT, serial TEXT, device_type TEXT,
                id_weight INT DEFAULT 1, name_weight INT DEFAULT 1, vendor_weight INT DEFAULT 1, hw_weight INT DEFAULT 1, sw_weight INT DEFAULT 1, productID_weight INT DEFAULT 1, serial_weight INT DEFAULT 1, device_type_weight INT DEFAULT 1,
                timestamp TIME, updated BOOLEAN )
                            ''')
                ## mac=0, protocol=1, ip=2
                ## id=3, name=4, vendor=5, hw=6, sw=7, productid=8, serial=9, device_type=10
                ## id_w=11, name_w=12, vendor_w=13, hw_w=14, sw_w=15, product_w=16, serial_w=17, device_w=18,
                ## time=19, updated=20
        self.cursor.execute('CREATE INDEX IF NOT EXISTS idx_mac ON endpoints (mac)')
        self.connection.commit()
        logger.debug('create endpoints DB - complete')
    
    def update_db_list(self, values_list):
        global records_created, records_updated                 # Values of DB record updates for efficiency tracking
        mac = values_list[0]                                    
        self.cursor.execute('SELECT * FROM endpoints WHERE mac = ?', (mac,))
        existing_record = self.cursor.fetchone()                # Grab any matching entry based on MAC
        if existing_record:
            update_values = {}                                  # Temp array to hold any new values
            if not existing_record[1] and values_list[1]:
                update_values['protocol'] = values_list[1]
            elif existing_record[1] != values_list[1]:          # Append new protocol to existing protocol string in DB
                if str(values_list[1]) not in str(existing_record[1]):
                    update_values['protocol'] = f'{existing_record[1]}, {values_list[1]}'
            if not existing_record[2] and values_list[2]:
                update_values['ip'] = values_list[2]
            elif existing_record[2] != values_list[2]:
                update_values['ip'] = values_list[2]            # Overwrite the IP address w/ newer data
            if (not existing_record[3] and values_list[3]) or (existing_record[3] != values_list[3] and int(existing_record[11]) < int(values_list[11])):
                update_values['id'] = values_list[3]            #If existing ID value is less weight than new value, update record values
                update_values['id_weight'] = values_list[11]
            if (not existing_record[4] and values_list[4]) or (existing_record[4] != values_list[4] and int(existing_record[12]) < int(values_list[12])):
                update_values['name'] = values_list[4]
                update_values['name_weight'] = values_list[12]
            if (not existing_record[5] and values_list[5]) or (existing_record[5] != values_list[5] and int(existing_record[13]) < int(values_list[13])):
                update_values['vendor'] = values_list[5]
                update_values['vendor_weight'] = values_list[13]
            if (not existing_record[6] and values_list[6]) or (existing_record[6] != values_list[6] and int(existing_record[14]) < int(values_list[14])):
                update_values['hw'] = values_list[6]
                update_values['hw_weight'] = values_list[14]
            if (not existing_record[7] and values_list[7]) or (existing_record[7] != values_list[7] and int(existing_record[15]) < int(values_list[15])):
                update_values['sw'] = values_list[7]
                update_values['sw_weight'] = values_list[15]
            if (not existing_record[8] and values_list[8]) or (existing_record[8] != values_list[8] and int(existing_record[16]) < int(values_list[16])):
                update_values['productID'] = values_list[8]
                update_values['productID_weight'] = values_list[16]
            if (not existing_record[9] and values_list[9]) or (existing_record[9] != values_list[9] and int(existing_record[17]) < int(values_list[17])):
                update_values['serial'] = values_list[9]
                update_values['serial_weight'] = values_list[17]
            if (not existing_record[10] and values_list[10]) or (existing_record[10] != values_list[10] and int(existing_record[18]) < int(values_list[18])):
                update_values['device_type'] = values_list[10]
                update_values['device_type_weight'] = values_list[18]

            # Update the record with new values by appending DB fields to SQL query based on values within the 'update_values' array
            if update_values:
                update_query = ', '.join([f"{field} = ?" for field in update_values.keys()])
                update_query += ', updated = ?, timestamp = ? WHERE mac = ?'
                # TO DO ## LOGIC FOR 
                update_values['updated'] = False
                update_values['timestamp'] = datetime.now().strftime("%H:%M:%S")
                self.cursor.execute(f"UPDATE endpoints SET {update_query}", (*update_values.values(), values_list[0]))
                records_updated += 1
                logger.debug(f'endpoint db record updated: {values_list[0]} - {values_list[1]} data')
                
        else:
            ## Insert the new endpoint into the database w/ relative fields, but 'FALSE' for 'updated' status
            self.cursor.execute('''
                INSERT INTO endpoints (
                    mac, protocol, ip, 
                    id, name, vendor, hw, sw, productID, serial, device_type, 
                    id_weight, name_weight, vendor_weight, hw_weight, sw_weight, productID_weight, serial_weight, device_type_weight, 
                    updated, timestamp
                )
                VALUES (?, ?, ?, 
                        ?, ?, ?, ?, ?, ?, ?, ?, 
                        ?, ?, ?, ?, ?, ?, ?, ?, 
                        ?, ?)''',
                (values_list[0], values_list[1], values_list[2], 
                 values_list[3], values_list[4], values_list[5], values_list[6], values_list[7], values_list[8], values_list[9], values_list[10], 
                 values_list[11], values_list[12], values_list[13], values_list[14], values_list[15], values_list[16], values_list[17], values_list[18],
                 False, datetime.now().strftime("%H:%M:%S")))
            logger.debug(f'endpoint db record created: {values_list[0]} - {values_list[1]} data')
            records_created += 1

        self.connection.commit()

    ## Function for data validation (replace this with your logic)
    def view_all_entries(self):
        self.cursor.execute('SELECT * FROM endpoints')
        entries = self.cursor.fetchall()

        print("All Entries in the 'endpoints' table:")
        for entry in entries:
            print(entry)
    
    ## View all entries in local DB with records that have not been updated and return as object
    async def get_active_entries(self):
        self.cursor.execute('SELECT * FROM endpoints WHERE updated = 0')
        entries = self.cursor.fetchall()
        return entries

    ## Modify local DB record 'Update' status to 'True' to avoid being included in future updates until new data is available
    async def ise_endpoint_updated(self, mac):
        global ise_updates
        self.cursor.execute('UPDATE endpoints SET updated = 1 WHERE mac = ?', (mac,))
        ise_updates += 1

    def view_stats(self):
        logger.debug(f'Local DB records created: {records_created}')
        logger.debug(f'Local DB records updated: {records_updated}')
        logger.debug(f'Local DB records sent to ISE: {ise_updates}')

    def close_connection(self):
        self.connection.close()
